# SPDX-FileCopyrightText: 2012-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Global settings used by all scripts in this directory.
# XXX Before any use of the tools in this directory, please make a copy of this file
#     named "setting.py"
# XXX This is a template, most values should be OK, but some you’ll have to
#     edit (most probably, BLENDER_EXEC and SOURCE_DIR).


import json
import os
import re
import sys
import types

# Only do soft-dependency on `bpy` module, not real strong need for it currently.
try:
    import bpy
except ModuleNotFoundError:
    bpy = None

###############################################################################
# MISC
###############################################################################

# The languages defined in Blender.
LANGUAGES = (
    # ID, UI English label, ISO code.
    (0, "Automatic", "DEFAULT"),
    (1, "English (US)", "en_US"),
    (2, "Japanese - 日本語", "ja_JP"),
    (3, "Dutch - Nederlands", "nl_NL"),
    (4, "Italian - Italiano", "it_IT"),
    (5, "German - Deutsch", "de_DE"),
    (6, "Finnish - Suomi", "fi_FI"),
    (7, "Swedish - Svenska", "sv_SE"),
    (8, "French - Français", "fr_FR"),
    (9, "Spanish - Español", "es"),
    (10, "Catalan - Català", "ca_AD"),
    (11, "Czech - Čeština", "cs_CZ"),
    (12, "Portuguese (Portugal) - Português europeu", "pt_PT"),
    (13, "Chinese (Simplified) - 简体中文", "zh_HANS"),
    (14, "Chinese (Traditional) - 繁體中文", "zh_HANT"),
    (15, "Russian - Русский", "ru_RU"),
    (16, "Croatian - Hrvatski", "hr_HR"),
    (17, "Serbian (Cyrillic) - Српски", "sr_RS"),
    (18, "Ukrainian - Українська", "uk_UA"),
    (19, "Polish - Polski", "pl_PL"),
    (20, "Romanian - Român", "ro_RO"),
    # Using the utf8 flipped form of Arabic (العربية).
    (21, "Arabic - ﺔﻴﺑﺮﻌﻟﺍ", "ar_EG"),
    (22, "Bulgarian - Български", "bg_BG"),
    (23, "Greek - Ελληνικά", "el_GR"),
    (24, "Korean - 한국어", "ko_KR"),
    (25, "Nepali - नेपाली", "ne_NP"),
    # Using the utf8 flipped form of Persian (فارسی).
    (26, "Persian - ﯽﺳﺭﺎﻓ", "fa_IR"),
    (27, "Indonesian - Bahasa indonesia", "id_ID"),
    (28, "Serbian (Latin) - Srpski latinica", "sr_RS@latin"),
    (29, "Kyrgyz - Кыргыз тили", "ky_KG"),
    (30, "Turkish - Türkçe", "tr_TR"),
    (31, "Hungarian - Magyar", "hu_HU"),
    (32, "Portuguese (Brazil) - Português brasileiro", "pt_BR"),
    # Using the utf8 flipped form of Hebrew (עִבְרִית)).
    (33, "Hebrew - תירִבְעִ", "he_IL"),
    (34, "Estonian - Eesti keel", "et_EE"),
    (35, "Esperanto - Esperanto", "eo"),
    # 36 is free, used to be 'Spanish from Spain' (`es_ES`).
    (37, "Amharic - አማርኛ", "am_ET"),
    (38, "Uzbek (Latin) - Oʻzbek", "uz_UZ@latin"),
    (39, "Uzbek (Cyrillic) - Ўзбек", "uz_UZ@cyrillic"),
    (40, "Hindi - हिन्दी", "hi_IN"),
    (41, "Vietnamese - Tiếng Việt", "vi_VN"),
    (42, "Basque - Euskara", "eu_EU"),
    (43, "Hausa - Hausa", "ha"),
    (44, "Kazakh - Қазақша", "kk_KZ"),
    (45, "Abkhaz - Аԥсуа бызшәа", "ab"),
    (46, "Thai - ภาษาไทย", "th_TH"),
    (47, "Slovak - Slovenčina", "sk_SK"),
    (48, "Georgian - ქართული", "ka"),
    (49, "Tamil - தமிழ்", "ta"),
    (50, "Khmer - ខ្មែរ", "km"),
    (51, "Swahili - Kiswahili", "sw"),
    (52, "Belarusian - беларуску", "be"),
    (53, "Danish - Dansk", "da"),
    (54, "Slovenian - Slovenščina", "sl"),
    # Using the utf8 flipped form of Urdu (اُردُو).
    (55, "Urdu - وُدرُا", "ur"),
    (56, "Lithuanian - Lietuviškai", "lt"),
    (57, "English (UK)", "en_GB"),
    (58, "Malayalam - മലയാളം", "ml"),
)

# Default context, in py (keep in sync with `BLT_translation.hh`)!
if bpy is not None:
    assert bpy.app.translations.contexts.default == "*"
DEFAULT_CONTEXT = "*"

# Name of language file used by Blender to generate translations' menu.
LANGUAGES_FILE = "languages"

# The minimum level of completeness for a `.po` file to be imported from
# the working repository to the Blender one, as a percentage.
IMPORT_MIN_LEVEL = 0.0

# Languages in the working repository that should not be imported in the Blender one currently...
IMPORT_LANGUAGES_SKIP = {
    'am_ET', 'et_EE', 'uz_UZ@latin', 'uz_UZ@cyrillic', 'kk_KZ',
}

# Languages that need RTL pre-processing.
IMPORT_LANGUAGES_RTL = {
    'ar_EG', 'fa_IR', 'he_IL', 'ur',
}

# The comment prefix used in generated `messages.txt` file.
MSG_COMMENT_PREFIX = "#~ "

# The comment prefix used in generated `messages.txt` file.
MSG_CONTEXT_PREFIX = "MSGCTXT:"

# The default comment prefix used in po's.
PO_COMMENT_PREFIX = "# "

# The comment prefix used to mark sources of msgids, in po's.
PO_COMMENT_PREFIX_SOURCE = "#: "

# The comment prefix used to mark sources of msgids, in po's.
PO_COMMENT_PREFIX_SOURCE_CUSTOM = "#. :src: "

# The general "generated" comment prefix, in po's.
PO_COMMENT_PREFIX_GENERATED = "#. "

# The comment prefix used to comment entries in po's.
PO_COMMENT_PREFIX_MSG = "#~ "

# The comment prefix used to mark fuzzy msgids, in po's.
PO_COMMENT_FUZZY = "#, fuzzy"

# The prefix used to define context, in po's.
PO_MSGCTXT = "msgctxt "

# The prefix used to define msgid, in po's.
PO_MSGID = "msgid "

# The prefix used to define msgstr, in po's.
PO_MSGSTR = "msgstr "

# The 'header' key of po files.
PO_HEADER_KEY = (DEFAULT_CONTEXT, "")

PO_HEADER_MSGSTR = (
    "Project-Id-Version: {blender_ver} ({blender_hash})\\n\n"
    "Report-Msgid-Bugs-To: \\n\n"
    "POT-Creation-Date: {time}\\n\n"
    "PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n\n"
    "Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n\n"
    "Language-Team: LANGUAGE <LL@li.org>\\n\n"
    "Language: {uid}\\n\n"
    "MIME-Version: 1.0\\n\n"
    "Content-Type: text/plain; charset=UTF-8\\n\n"
    "Content-Transfer-Encoding: 8bit\n"
)
PO_HEADER_COMMENT_COPYRIGHT = (
    "# Blender's translation file (po format).\n"
    "# Copyright (C) {year} The Blender Authors.\n"
    "# This file is distributed under the same license as the Blender package.\n"
    "#\n"
)
PO_HEADER_COMMENT = (
    "# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.\n"
    "#"
)

TEMPLATE_ISO_ID = "__TEMPLATE__"

# Num buttons report their label with a trailing ': '...
NUM_BUTTON_SUFFIX = ": "

# Undocumented operator placeholder string.
UNDOC_OPS_STR = "(undocumented operator)"

# The gettext domain.
DOMAIN = "blender"

# Our own "gettext" stuff.
# File type (ext) to parse.
PYGETTEXT_ALLOWED_EXTS = {".c", ".cc", ".cpp", ".cxx", ".hh", ".hpp", ".hxx", ".h"}

# Max number of contexts into a BLT_I18N_MSGID_MULTI_CTXT macro...
PYGETTEXT_MAX_MULTI_CTXT = 16

# Where to search contexts definitions, relative to SOURCE_DIR (defined below).
PYGETTEXT_CONTEXTS_DEFSRC = os.path.join("source", "blender", "blentranslation", "BLT_translation.h")

# Regex to extract contexts defined in BLT_translation.h
# XXX Not full-proof, but should be enough here!
PYGETTEXT_CONTEXTS = "#define\\s+(BLT_I18NCONTEXT_[A-Z_0-9]+)\\s+\"([^\"]*)\""

# autopep8: off

# Keywords' regex.
# XXX Most unfortunately, we can't use named back-references inside character sets,
#     which makes the REGEXES even more twisty... :/
_str_base = (
    # Match void string
    "(?P<{_}1>[\"'])(?P={_}1)"  # Get opening quote (' or "), and closing immediately.
    "|"
    # Or match non-void string
    "(?P<{_}2>[\"'])"  # Get opening quote (' or ").
        "(?{capt}(?:"
            # This one is for crazy things like "hi \\\\\" folks!"...
            r"(?:(?!<\\)(?:\\\\)*\\(?=(?P={_}2)))|"
            # The most common case.
            ".(?!(?P={_}2))"
        ")*.)"  # Don't forget the last char!
    "(?P={_}2)"  # And closing quote.
)
str_clean_re = _str_base.format(_="g", capt="P<clean>")
_inbetween_str_re = (
    # XXX Strings may have comments between their pieces too, not only spaces!
    r"(?:\s*(?:"
        # A C comment
        r"/\*.*(?!\*/).\*/|"
        # Or a C++ one!
        r"//[^\n]*\n|"
        # Or some #defined value (like `BLI_STR_UTF8_BLACK_RIGHT_POINTING_SMALL_TRIANGLE`)
        # NOTE: This should be avoided at all cost, as it will simply make translation lookup fail.
        r"[ a-zA-Z0-9_]*"
    # And we are done!
    r")?)*"
)
# Here we have to consider two different cases (empty string and other).
_str_whole_re = (
    _str_base.format(_="{_}1_", capt=":") +
    # Optional loop start, this handles "split" strings...
    "(?:(?<=[\"'])" + _inbetween_str_re + "(?=[\"'])(?:"
        + _str_base.format(_="{_}2_", capt=":") +
    # End of loop.
    "))*"
)
_ctxt_re_gen = lambda uid: (
    r"(?P<ctxt_raw{uid}>(?:".format(uid=uid) +
    _str_whole_re.format(_="_ctxt{uid}".format(uid=uid)) +
    r")|(?:[A-Z_0-9]+))"
)
_ctxt_re = _ctxt_re_gen("")
_msg_re = r"(?P<msg_raw>" + _str_whole_re.format(_="_msg") + r")"

class PyGettextKeyword:
    def __init__(self, re_expr, context_override=None):
        self.re_expr = re_expr
        self.context_override = context_override

        self.re = re.compile(re_expr)
        self.search = self.re.search

PYGETTEXT_KEYWORDS = (() +
    tuple(PyGettextKeyword((r"{}\(\s*" + _msg_re + r"\s*\)").format(it))
         for it in ("IFACE_", "TIP_", "RPT_", "DATA_", "N_")) +

    tuple(PyGettextKeyword((r"{}\(\s*" + _ctxt_re + r"\s*,\s*" + _msg_re + r"\s*\)").format(it))
          for it in ("CTX_IFACE_", "CTX_TIP_", "CTX_RPT_", "CTX_DATA_", "CTX_N_")) +

    tuple(PyGettextKeyword(("{}\\((?:[^\"',]+,){{1,2}}\\s*" + _msg_re + r"\s*(?:\)|,)").format(it))
          for it in ("BKE_report", "BKE_reportf", "BKE_reports_prepend", "BKE_reports_prependf",
                     "CTX_wm_operator_poll_msg_set", "WM_global_report", "WM_global_reportf",
                     "UI_but_disable")) +

    # ED_undo_push() is used in Undo History menu and has the "Operator" context, same as operators.
    tuple(PyGettextKeyword(("{}\\((?:[^\"',]+,)\\s*" + _msg_re + r"\s*(?:\))").format(it),
                           context_override='BLT_I18NCONTEXT_OPERATOR_DEFAULT')
          for it in ("ED_undo_push", "ED_undo_grouped_push")) +

    # bmesh operator errors
    tuple(PyGettextKeyword(("{}\\((?:[^\"',]+,){{3}}\\s*" + _msg_re + r"\s*\)").format(it))
          for it in ("BMO_error_raise",)) +

    # Modifier errors
    tuple(PyGettextKeyword(("{}\\((?:[^\"',]+,){{2}}\\s*" + _msg_re + r"\s*(?:\)|,)").format(it))
          for it in ("BKE_modifier_set_error",)) +

    # Window manager job names.
    tuple(PyGettextKeyword(("{}\\((?:[^\"',]+,){{3}}\\s*" + _msg_re + r"\s*,").format(it))
          for it in ("WM_jobs_get",)) +

    # Compositor and EEVEE messages.
    # Ends either with `)` (function call close), or `,` when there are extra formatting parameters.
    tuple(PyGettextKeyword((r"{}\(\s*" + _msg_re + r"\s*(?:\)|,)").format(it))
          for it in ("set_info_message", "info_append_i18n")) +

    # This one is a tad more risky, but in practice would not expect a name/uid string parameter
    # (the second one in those functions) to ever have a comma in it, so think this is fine.
    tuple(PyGettextKeyword(("{}\\((?:[^,]+,){{2}}\\s*" + _msg_re + r"\s*(?:\)|,)").format(it))
          for it in ("modifier_subpanel_register", "gpencil_modifier_subpanel_register")) +

    # Node socket declarations: context-less names.
    tuple(PyGettextKeyword((r"\.{}(?:<decl::.*?>\(|[^,]+,)\s*" + _msg_re + r"(?:,[^),]+)*\s*\)"
           r"(?![^;]*\.translation_context\()").format(it))
          for it in ("add_input", "add_output")) +

    # Node socket declarations: names with contexts
    tuple(PyGettextKeyword((r"\.{}(?:<decl::.*?>\(|[^,]+,)\s*" + _msg_re +
           r"[^;]*\.translation_context\(\s*" + _ctxt_re + r"\s*\)").format(it))
          for it in ("add_input", "add_output")) +

    # Node socket declarations: description and error messages
    tuple(PyGettextKeyword((r"\.{}\(\s*" + _msg_re + r"\s*\)").format(it))
          for it in ("description", "error_message_add")) +

    # Node socket panels and labels from declarations: context-less names
    tuple(PyGettextKeyword((r"\.{}\(\s*" + _msg_re +
           r"\s*\)(?![^;]*\.translation_context\()[^;]*;").format(it))
          for it in ("short_label", "add_panel",)) +

    # Node socket panels and labels from declarations: names with contexts
    tuple(PyGettextKeyword((r"\.{}\(\s*" + _msg_re + r"[^;]*\.translation_context\(\s*" +
           _ctxt_re + r"\s*\)").format(it))
          for it in ("short_label", "add_panel",)) +

    # Dynamic node socket labels
    tuple(PyGettextKeyword((r"{}\(\s*[^,]+,\s*" + _msg_re + r"\s*\)").format(it))
          for it in ("node_sock_label",)) +

    # Geometry Nodes field inputs
    (PyGettextKeyword(r"FieldInput\(CPPType::get<.*?>\(\),\s*" + _msg_re + r"\s*\)"),) +

    # bUnitDef unit names
    (PyGettextKeyword(r"/\*name_display\*/\s*" + _msg_re + r"\s*,"),) +

    tuple(PyGettextKeyword(
        (r"{}\(\s*"
         + _msg_re
         + r"\s*,\s*(?:"
         + r"\s*,\s*)?(?:".join(_ctxt_re_gen(i) for i in range(PYGETTEXT_MAX_MULTI_CTXT))
         + r")?\s*,?\s*\)"
         ).format(it)
    ) for it in ("BLT_I18N_MSGID_MULTI_CTXT",))
)

# autopep8: on


# Check printf mismatches between msgid and msgstr.
CHECK_PRINTF_FORMAT = (
    r"(?!<%)(?:%%)*%"          # Beginning, with handling for crazy things like '%%%%%s'
    r"[-+#0]?"                 # Flags (note: do not add the ' ' (space) flag here, generates too much false positives!)
    r"(?:\*|[0-9]+)?"          # Width
    r"(?:\.(?:\*|[0-9]+))?"    # Precision
    r"(?:[hljztL]|hh|ll)?"     # Length
    r"[tldiuoxXfFeEgGaAcspn]"  # Specifiers (note we have Blender-specific %t and %l ones too)
)

# Should po parser warn when finding a first letter not capitalized?
WARN_MSGID_NOT_CAPITALIZED = True

# Strings that should not raise above warning!
WARN_MSGID_NOT_CAPITALIZED_ALLOWED = {
    "",                              # Simplifies things... :p
    "ac3",
    "along X",
    "along Y",
    "along Z",
    "along %s X",
    "along %s Y",
    "along %s Z",
    "along local Z",
    "arccos(A)",
    "arcsin(A)",
    "arctan(A)",
    "ascii",
    "author",                        # Addons' field. :/
    "bItasc",
    "blender.org",
    "bytes",
    "color_index is invalid",
    "cos(A)",
    "cosh(A)",
    "dB",                            # dB audio power unit.
    "dbl-",                          # Compacted for 'double', for keymap items.
    "description",                   # Addons' field. :/
    "dx",
    "fBM",
    "flac",
    "fps: %.2f",
    "fps: %i",
    "gimbal",
    "global",
    "glTF 2.0 (.glb/.gltf)",
    "glTF Binary (.glb)",
    "glTF Embedded (.gltf)",
    "glTF Material Output",
    "glTF Original PBR data",
    "glTF Separate (.gltf + .bin + textures)",
    "gltfpack",
    "glTFpack file path",
    "invoke() needs to be called before execute()",
    "iScale",
    "iso-8859-15",
    "iTaSC",
    "iTaSC parameters",
    "kb",
    "local",
    "location",                      # Addons' field. :/
    "locking %s X",
    "locking %s Y",
    "locking %s Z",
    "mkv",
    "mm",
    "mp2",
    "mp3",
    "normal",
    "ogg",
    "oneAPI",
    "p0",
    "parent_index should not be less than -1: %d",
    "parent_index (%d) should be less than the number of bone collections (%d)",
    "px",
    "re",
    "res",
    "rv",
    "seconds",
    "sin(A)",
    "sin(x) / x",
    "sinh(A)",
    "sqrt(x*x+y*y+z*z)",
    "sRGB",
    "sRGB display space",
    "sRGB display space with Filmic view transform",
    "sRGB IEC 61966-2-1 compound (piece-wise) encoding",
    "tan(A)",
    "tanh(A)",
    "utf-8",
    "uv_on_emitter() requires a modifier from an evaluated object",
    "var",
    "vBVH",
    "view",
    "wav",
    "wmOwnerID '%s' not in workspace '%s'",
    "y",
    "y = (Ax + B)",
    # ID plural names, defined in IDTypeInfo.
    "armatures",
    "brushes",
    "cache_files",
    "cameras",
    "collections",
    "curves",
    "fonts",
    "grease_pencils",
    "hair_curves",
    "ipos",
    "lattices",
    "libraries",
    "lightprobes",
    "lights",
    "linestyles",
    "link_placeholders",
    "masks",
    "metaballs",
    "materials",
    "meshes",
    "movieclips",
    "node_groups",
    "objects",
    "paint_curves",
    "palettes",
    "particles",
    "pointclouds",
    "screens",
    "shape_keys",
    "sounds",
    "speakers",
    "texts",
    "textures",
    "volumes",
    "window_managers",
    "workspaces",
    "worlds",
    # Sub-strings.
    "all",
    "all and invert unselected",
    "and AMD driver version %s or newer",
    "and AMD Radeon Pro %s driver or newer",
    "and NVIDIA driver version %s or newer",
    "and Windows driver version %s or newer",
    "available with",
    "brown fox",
    "can't save image while rendering",
    "category",
    "constructive modifier",
    "cursor",
    "custom",
    "custom matrix",
    "custom orientation",
    "drag-",
    "edge data",
    "exp(A)",
    "expected a timeline/animation area to be active",
    "expected a view3d region",
    "expected a view3d region & editcurve",
    "expected a view3d region & editmesh",
    "face data",
    "gimbal",
    "global",
    "glTF Settings",
    "image file not found",
    "image format is read-only",
    "image path can't be written to",
    "in %i days",
    "in %i hours",
    "in %i minutes",
    "in memory to enable editing!",
    "in the asset shelf.",
    "insufficient content",
    "into",
    "jumps over",
    "left",
    "local",
    "matrices", "no matrices",
    "multi-res modifier",
    "name",
    "non-triangle face",
    "normal",
    "on {:%Y-%m-%d}",
    "or AMD with macOS %s or newer",
    "parent",
    "performance impact!",
    "positions", "no positions",
    "read",
    "remove",
    "right",
    "selected",
    "selected and lock unselected",
    "selected and unlock unselected",
    "screen",
    "the lazy dog",
    "this legacy pose library to pose assets",
    "to the top level of the tree",
    "unable to load movie clip",
    "unable to load text",
    "unable to open the file",
    "unknown error reading file",
    "unknown error statting file",
    "unknown error writing file",
    "unselected",
    "unsupported font format",
    "unsupported format",
    "unsupported image format",
    "unsupported movie clip format",
    "untitled",
    "vertex data",
    "verts only",
    "view",
    "virtual parents",
    "which was replaced by the Asset Browser",
    "within seconds",
    "write",
}
WARN_MSGID_NOT_CAPITALIZED_ALLOWED |= set(lng[2] for lng in LANGUAGES)

WARN_MSGID_END_POINT_ALLOWED = {
    "Cannot figure out which object this bone belongs to.",
    "Circle|Alt .",
    "Float Neg. Exp.",
    "Max Ext.",
    "Newer graphics drivers may be available to improve Blender support.",
    "Not assigned to any bone collection.",
    "Numpad .",
    "Pad.",
    "Please file a bug report.",
    "    RNA Path: bpy.types.",
    "Temp. Diff.",
    "Temperature Diff.",
    "The program will now close.",
    "Your graphics card or driver has limited support. It may work, but with issues.",
    "Your graphics card or driver is not supported.",
    "Invalid surface UVs on %d curves.",
    "The pose library moved.",
    "in the asset shelf.",
    "Remove, local files not found.",
    "Remove all files in \"{}\".",
    "Remove, keeping local files.",
}

PARSER_CACHE_HASH = 'sha1'

PARSER_TEMPLATE_ID = "__POT__"
PARSER_PY_ID = "__PY__"

PARSER_PY_MARKER_BEGIN = "\n# ##### BEGIN AUTOGENERATED I18N SECTION #####\n"
PARSER_PY_MARKER_END = "\n# ##### END AUTOGENERATED I18N SECTION #####\n"

PARSER_MAX_FILE_SIZE = 2 ** 24  # in bytes, i.e. 16 Mb.

###############################################################################
# PATHS
###############################################################################

# The Python3 executable.You’ll likely have to edit it in your user_settings.py
# if you’re under Windows.
PYTHON3_EXEC = "python3"

# The Blender executable!
# This is just an example, you’ll have to edit it in your user_settings.py!
BLENDER_EXEC = os.path.abspath(os.path.join("foo", "bar", "blender"))
# check for blender.bin
if not os.path.exists(BLENDER_EXEC):
    if os.path.exists(BLENDER_EXEC + ".bin"):
        BLENDER_EXEC = BLENDER_EXEC + ".bin"

# The gettext msgfmt "compiler". You’ll likely have to edit it in your user_settings.py if you’re under Windows.
GETTEXT_MSGFMT_EXECUTABLE = "msgfmt"

# The FriBidi C compiled library (.so under Linux, `.dll` under windows...).
# You’ll likely have to edit it in your `user_settings.py` if you’re under Windows., e.g. using the included one:
# `FRIBIDI_LIB = os.path.join(TOOLS_DIR, "libfribidi.dll")`
FRIBIDI_LIB = "libfribidi.so.0"

# The name of the (currently empty) file that must be present in a po's directory to enable RTL-preprocess.
RTL_PREPROCESS_FILE = "is_rtl"

# The Blender source root path.
# This is just an example, you’ll have to override it in your user_settings.py!
SOURCE_DIR = os.path.abspath(os.path.join("blender"))

# The bf-translation repository (you'll have to override this in your user_settings.py).
I18N_DIR = os.path.abspath(os.path.join("i18n"))

# The 'work' path to PO files (relative to I18N_DIR).
REL_WORK_DIR = os.path.join("")


# The path to the Blender translation directory (relative to SOURCE_DIR).
REL_BLENDER_I18N_DIR = os.path.join("locale")

# The /po path of the Blender translation directory (relative to REL_BLENDER_I18N_DIR).
REL_BLENDER_I18N_PO_DIR = os.path.join("po")


# The Blender source path to check for i18n macros (relative to SOURCE_DIR).
REL_POTFILES_SOURCE_DIR = os.path.join("source")

# Where to search for preset names (relative to SOURCE_DIR).
REL_PRESETS_DIR = os.path.join("scripts", "presets")

# Where to search for templates (relative to SOURCE_DIR).
REL_TEMPLATES_DIR = os.path.join("scripts", "startup", "bl_app_templates_system")

# Name of the built-in asset catalog file.
ASSET_CATALOG_FILE = "blender_assets.cats.txt"

# The template messages file (relative to I18N_DIR).
REL_FILE_NAME_POT = os.path.join(REL_WORK_DIR, DOMAIN + ".pot")


# Mo path generator for a given language (relative to any "locale" directory).
MO_PATH_ROOT_RELATIVE = os.path.join("locale")
MO_PATH_TEMPLATE_RELATIVE = os.path.join(MO_PATH_ROOT_RELATIVE, "{}", "LC_MESSAGES")

# Mo file name.
MO_FILE_NAME = DOMAIN + ".mo"

# Where to search for py files that may contain ui strings (relative to one of the 'resource_path' of Blender).
CUSTOM_PY_UI_FILES = [
    os.path.join("scripts", "startup", "bl_ui"),
    os.path.join("scripts", "startup", "bl_operators"),
    os.path.join("scripts", "modules", "rna_prop_ui.py"),
    os.path.join("scripts", "modules", "rna_keymap_ui.py"),
    os.path.join("scripts", "modules", "_bpy_types.py"),
    os.path.join("scripts", "presets", "keyconfig"),
]

# An optional text file listing files to force include/exclude from py_xgettext process.
SRC_POTFILES = ""

# A cache storing validated msgids, to avoid re-spellchecking them.
SPELL_CACHE = os.path.join("/tmp", ".spell_cache")

# Threshold defining whether a new msgid is similar enough with an old one to reuse its translation...
SIMILAR_MSGID_THRESHOLD = 0.75

# Additional import paths to add to `sys.path` (';' separated)...
INTERN_PY_SYS_PATHS = ""

# Custom override settings must be one directory above i18n tools itself!
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
try:
    from bl_i18n_settings_override import *
except ImportError:  # If no i18n_override_settings available, it’s no error!
    pass

# Override with custom user settings, if available.
try:
    from settings_user import *
except ImportError:  # If no user_settings available, it’s no error!
    pass


for p in set(INTERN_PY_SYS_PATHS.split(";")):
    if p:
        sys.path.append(p)


# The settings class itself!
def _do_get(ref, path):
    return os.path.normpath(os.path.join(ref, path))


def _do_set(ref, path):
    path = os.path.normpath(path)
    # If given path is absolute, make it relative to current ref one (else we consider it is already the case!)
    if os.path.isabs(path):
        # can't always find the relative path (between drive letters on windows)
        try:
            return os.path.relpath(path, ref)
        except ValueError:
            pass
    return path


def _gen_get_set_path(ref, name):
    def _get(self):
        return _do_get(getattr(self, ref), getattr(self, name))

    def _set(self, value):
        setattr(self, name, _do_set(getattr(self, ref), value))
    return _get, _set


def _check_valid_data(uid, val):
    return not uid.startswith("_") and type(val) not in tuple(types.__dict__.values()) + (type,)


class I18nSettings:
    """
    Class allowing persistence of our settings!
    Saved in JSON format, so settings should be JSON'able objects!
    """
    _settings = None

    def __new__(cls, *args, **kwargs):
        # Addon preferences are singleton by definition, so is this class!
        if not I18nSettings._settings:
            cls._settings = super(I18nSettings, cls).__new__(cls)
            cls._settings.__dict__ = {uid: val for uid, val in globals().items() if _check_valid_data(uid, val)}
        return I18nSettings._settings

    def __getstate__(self):
        return self.to_dict()

    def __setstate__(self, mapping):
        return self.from_dict(mapping)

    def from_dict(self, mapping):
        # Special case... :/
        if "INTERN_PY_SYS_PATHS" in mapping:
            self.PY_SYS_PATHS = mapping["INTERN_PY_SYS_PATHS"]
        self.__dict__.update(mapping)

    def to_dict(self):
        glob = globals()
        return {uid: val for uid, val in self.__dict__.items() if _check_valid_data(uid, val) and uid in glob}

    def from_json(self, string):
        self.from_dict(dict(json.loads(string)))

    def to_json(self):
        # Only save the diff from default i18n_settings!
        glob = globals()
        export_dict = {
            uid: val for uid, val in self.__dict__.items()
            if _check_valid_data(uid, val) and glob.get(uid) != val
        }
        return json.dumps(export_dict)

    def load(self, fname, reset=False):
        reset = reset or fname is None
        if reset:
            self.__dict__ = {uid: data for uid, data in globals().items() if not uid.startswith("_")}
        if fname is None:
            return
        if isinstance(fname, str):
            if not os.path.isfile(fname):
                # Assume it is already real JSON string.
                self.from_json(fname)
                return
            with open(fname, encoding="utf8") as f:
                self.from_json(f.read())
        # Else assume fname is already a file(like) object!
        else:
            self.from_json(fname.read())

    def save(self, fname):
        if isinstance(fname, str):
            with open(fname, 'w', encoding="utf8") as f:
                f.write(self.to_json())
        # Else assume fname is already a file(like) object!
        else:
            fname.write(self.to_json())

    WORK_DIR = property(*(_gen_get_set_path("I18N_DIR", "REL_WORK_DIR")))
    BLENDER_I18N_ROOT = property(*(_gen_get_set_path("SOURCE_DIR", "REL_BLENDER_I18N_DIR")))
    BLENDER_I18N_PO_DIR = property(*(_gen_get_set_path("BLENDER_I18N_ROOT", "REL_BLENDER_I18N_PO_DIR")))
    POTFILES_SOURCE_DIR = property(*(_gen_get_set_path("SOURCE_DIR", "REL_POTFILES_SOURCE_DIR")))
    PRESETS_DIR = property(*(_gen_get_set_path("SOURCE_DIR", "REL_PRESETS_DIR")))
    TEMPLATES_DIR = property(*(_gen_get_set_path("SOURCE_DIR", "REL_TEMPLATES_DIR")))
    FILE_NAME_POT = property(*(_gen_get_set_path("I18N_DIR", "REL_FILE_NAME_POT")))

    def _get_py_sys_paths(self):
        return self.INTERN_PY_SYS_PATHS

    def _set_py_sys_paths(self, val):
        old_paths = set(self.INTERN_PY_SYS_PATHS.split(";")) - {""}
        new_paths = set(val.split(";")) - {""}
        for p in old_paths - new_paths:
            if p in sys.path:
                sys.path.remove(p)
        for p in new_paths - old_paths:
            sys.path.append(p)
        self.INTERN_PY_SYS_PATHS = val
    PY_SYS_PATHS = property(_get_py_sys_paths, _set_py_sys_paths)



# SPDX-FileCopyrightText: 2011-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from bpy.app.translations import (
    pgettext_rpt as rpt_,
)
from bpy.types import (
    Operator,
)
from bpy.props import (
    BoolProperty,
    EnumProperty,
    StringProperty,
)


class SCENE_OT_freestyle_fill_range_by_selection(Operator):
    """Fill the Range Min/Max entries by the min/max distance between selected mesh objects and the source object """ \
        """(either a user-specified object or the active camera)"""
    bl_idname = "scene.freestyle_fill_range_by_selection"
    bl_label = "Fill Range by Selection"
    bl_options = {'INTERNAL'}

    type: EnumProperty(
        name="Type", description="Type of the modifier to work on",
        items=(
            ('COLOR', "Color", "Color modifier type"),
            ('ALPHA', "Alpha", "Alpha modifier type"),
            ('THICKNESS', "Thickness", "Thickness modifier type"),
        ),
    )
    name: StringProperty(
        name="Name",
        description="Name of the modifier to work on",
    )

    @classmethod
    def poll(cls, context):
        view_layer = context.view_layer
        return view_layer and view_layer.freestyle_settings.linesets.active

    def execute(self, context):
        import sys

        scene = context.scene
        view_layer = context.view_layer
        lineset = view_layer.freestyle_settings.linesets.active
        linestyle = lineset.linestyle
        # Find the modifier to work on
        if self.type == 'COLOR':
            m = linestyle.color_modifiers[self.name]
        elif self.type == 'ALPHA':
            m = linestyle.alpha_modifiers[self.name]
        else:
            m = linestyle.thickness_modifiers[self.name]
        # Find the reference object
        if m.type == 'DISTANCE_FROM_CAMERA':
            ref = scene.camera
            if ref is None:
                self.report({'ERROR'}, "No active camera in the scene")
                return {'CANCELLED'}
            matrix_to_camera = ref.matrix_world.inverted()
        elif m.type == 'DISTANCE_FROM_OBJECT':
            if m.target is None:
                self.report({'ERROR'}, "Target object not specified")
                return {'CANCELLED'}
            ref = m.target
            target_location = ref.location
        else:
            self.report({'ERROR'}, rpt_("Unexpected modifier type: {:s}").format(m.type))
            return {'CANCELLED'}
        # Find selected vertices in edit-mesh.
        ob = context.active_object
        if ob.type == 'MESH' and ob.mode == 'EDIT' and ob.name != ref.name:
            bpy.ops.object.mode_set(mode='OBJECT')
            selected_verts = [v for v in ob.data.vertices if v.select]
            bpy.ops.object.mode_set(mode='EDIT')
            # Compute the min/max distance from the reference to mesh vertices
            min_dist = sys.float_info.max
            max_dist = -min_dist
            if m.type == 'DISTANCE_FROM_CAMERA':
                ob_to_cam = matrix_to_camera @ ob.matrix_world
                for vert in selected_verts:
                    # dist in the camera space
                    dist = (ob_to_cam @ vert.co).length
                    min_dist = min(dist, min_dist)
                    max_dist = max(dist, max_dist)
            elif m.type == 'DISTANCE_FROM_OBJECT':
                for vert in selected_verts:
                    # dist in the world space
                    dist = (ob.matrix_world @ vert.co - target_location).length
                    min_dist = min(dist, min_dist)
                    max_dist = max(dist, max_dist)
            # Fill the Range Min/Max entries with the computed distances
            m.range_min = min_dist
            m.range_max = max_dist
            return {'FINISHED'}
        # Find selected mesh objects
        selection = [ob for ob in scene.objects if ob.select_get() and ob.type == 'MESH' and ob.name != ref.name]
        if selection:
            # Compute the min/max distance from the reference to mesh vertices
            min_dist = sys.float_info.max
            max_dist = -min_dist
            if m.type == 'DISTANCE_FROM_CAMERA':
                for ob in selection:
                    ob_to_cam = matrix_to_camera @ ob.matrix_world
                    for vert in ob.data.vertices:
                        # dist in the camera space
                        dist = (ob_to_cam @ vert.co).length
                        min_dist = min(dist, min_dist)
                        max_dist = max(dist, max_dist)
            elif m.type == 'DISTANCE_FROM_OBJECT':
                for ob in selection:
                    for vert in ob.data.vertices:
                        # dist in the world space
                        dist = (ob.matrix_world @ vert.co - target_location).length
                        min_dist = min(dist, min_dist)
                        max_dist = max(dist, max_dist)
            # Fill the Range Min/Max entries with the computed distances
            m.range_min = min_dist
            m.range_max = max_dist
        return {'FINISHED'}


class SCENE_OT_freestyle_add_edge_marks_to_keying_set(Operator):
    """Add the data paths to the Freestyle Edge Mark property of selected edges to the active keying set"""
    bl_idname = "scene.freestyle_add_edge_marks_to_keying_set"
    bl_label = "Add Edge Marks to Keying Set"
    bl_options = {'UNDO'}

    @classmethod
    def poll(cls, context):
        ob = context.active_object
        return (ob and ob.type == 'MESH')

    def execute(self, context):
        # active keying set
        scene = context.scene
        ks = scene.keying_sets.active
        if ks is None:
            ks = scene.keying_sets.new(idname="FreestyleEdgeMarkKeyingSet", name="Freestyle Edge Mark Keying Set")
            ks.bl_description = ""
        # add data paths to the keying set
        ob = context.active_object
        ob_mode = ob.mode
        mesh = ob.data
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        for i, edge in enumerate(mesh.edges):
            if not edge.hide and edge.select:
                path = "attributes[\"freestyle_edge\"].data[{:d}].value".format(i)
                ks.paths.add(mesh, path, index=0)
        bpy.ops.object.mode_set(mode=ob_mode, toggle=False)
        return {'FINISHED'}


class SCENE_OT_freestyle_add_face_marks_to_keying_set(Operator):
    """Add the data paths to the Freestyle Face Mark property of selected polygons to the active keying set"""
    bl_idname = "scene.freestyle_add_face_marks_to_keying_set"
    bl_label = "Add Face Marks to Keying Set"
    bl_options = {'UNDO'}

    @classmethod
    def poll(cls, context):
        ob = context.active_object
        return (ob and ob.type == 'MESH')

    def execute(self, context):
        # active keying set
        scene = context.scene
        ks = scene.keying_sets.active
        if ks is None:
            ks = scene.keying_sets.new(idname="FreestyleFaceMarkKeyingSet", name="Freestyle Face Mark Keying Set")
            ks.bl_description = ""
        # add data paths to the keying set
        ob = context.active_object
        ob_mode = ob.mode
        mesh = ob.data
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        for i, polygon in enumerate(mesh.polygons):
            if not polygon.hide and polygon.select:
                path = "attributes[\"freestyle_face\"].data[{:d}].value".format(i)
                ks.paths.add(mesh, path, index=0)
        bpy.ops.object.mode_set(mode=ob_mode, toggle=False)
        return {'FINISHED'}


class SCENE_OT_freestyle_module_open(Operator):
    """Open a style module file"""
    bl_idname = "scene.freestyle_module_open"
    bl_label = "Open Style Module File"
    bl_options = {'INTERNAL'}

    filepath: StringProperty(subtype='FILE_PATH')

    make_internal: BoolProperty(
        name="Make internal",
        description="Make module file internal after loading",
        default=True,
    )

    @classmethod
    def poll(cls, context):
        view_layer = context.view_layer
        return view_layer and view_layer.freestyle_settings.mode == 'SCRIPT'

    def invoke(self, context, _event):
        self.freestyle_module = context.freestyle_module
        wm = context.window_manager
        wm.fileselect_add(self)
        return {'RUNNING_MODAL'}

    def execute(self, _context):
        text = bpy.data.texts.load(self.filepath, internal=self.make_internal)
        self.freestyle_module.script = text
        return {'FINISHED'}


classes = (
    SCENE_OT_freestyle_add_edge_marks_to_keying_set,
    SCENE_OT_freestyle_add_face_marks_to_keying_set,
    SCENE_OT_freestyle_fill_range_by_selection,
    SCENE_OT_freestyle_module_open,
)



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from itertools import count

from ...utils.naming import make_derived_name
from ...utils.bones import align_bone_orientation
from ...utils.widgets_basic import create_circle_widget
from ...utils.widgets_special import create_neck_bend_widget, create_neck_tweak_widget
from ...utils.switch_parent import SwitchParentBuilder
from ...utils.misc import map_list

from ...base_rig import stage

from .spine_rigs import BaseHeadTailRig


class Rig(BaseHeadTailRig):
    """
    Head rig with long neck support and connect option.
    """

    use_connect_reverse = False
    min_chain_length = 1

    long_neck: bool
    has_neck: bool

    def initialize(self):
        super().initialize()

        self.long_neck = len(self.bones.org) > 3
        self.has_neck = len(self.bones.org) > 1

    ####################################################
    # BONES

    class CtrlBones(BaseHeadTailRig.CtrlBones):
        neck: str                      # Main neck control
        head: str                      # Main head control
        neck_bend: str                 # Extra neck bend control for long neck

    class MchBones(BaseHeadTailRig.MchBones):
        rot_neck: str                  # Main neck control parent for FK follow
        rot_head: str                  # Main head control parent for FK follow
        stretch: str                   # Long neck stretch helper
        ik: list[str]                  # Long neck IK system
        chain: list[str]               # Tweak parents

    bones: BaseHeadTailRig.ToplevelBones[
        list[str],
        'Rig.CtrlBones',
        'Rig.MchBones',
        list[str]
    ]

    ####################################################
    # Main control bones

    @stage.generate_bones
    def make_control_chain(self):
        orgs = self.bones.org
        ctrl = self.bones.ctrl

        if self.has_neck:
            ctrl.neck = self.make_neck_control_bone(orgs[0], 'neck', orgs[-1])

        ctrl.head = self.make_head_control_bone(orgs[-1], 'head')

        if self.long_neck:
            ctrl.neck_bend = self.make_neck_bend_control_bone(orgs[0], 'neck_bend', ctrl.neck)

        self.default_prop_bone = ctrl.head

    def make_neck_control_bone(self, org: str, name: str, org_head: str):
        name = self.copy_bone(org, name, parent=False)

        # Neck spans all neck bones (except head)
        self.get_bone(name).tail = self.get_bone(org_head).head

        return name

    def make_neck_bend_control_bone(self, org: str, name: str, neck: str):
        name = self.copy_bone(org, name, parent=False)
        neck_bend_eb = self.get_bone(name)

        # Neck pivot position
        neck_bones = self.bones.org
        if (len(neck_bones) - 1) % 2:     # odd num of neck bones (head excluded)
            center_bone = self.get_bone(neck_bones[int((len(neck_bones)) / 2) - 1])
            neck_bend_eb.head = (center_bone.head + center_bone.tail) / 2
        else:
            center_bone = self.get_bone(neck_bones[int((len(neck_bones) - 1) / 2) - 1])
            neck_bend_eb.head = center_bone.tail

        align_bone_orientation(self.obj, name, neck)
        neck_bend_eb.length = self.get_bone(neck).length / 2

        return name

    def make_head_control_bone(self, org: str, name: str):
        return self.copy_bone(org, name, parent=False)

    @stage.parent_bones
    def parent_control_chain(self):
        ctrl = self.bones.ctrl
        mch = self.bones.mch
        if self.has_neck:
            self.set_bone_parent(ctrl.neck, mch.rot_neck)
        self.set_bone_parent(ctrl.head, mch.rot_head)
        if self.long_neck:
            self.set_bone_parent(ctrl.neck_bend, mch.stretch)

    @stage.configure_bones
    def configure_control_chain(self):
        if self.has_neck:
            self.configure_control_bone(0, self.bones.ctrl.neck, self.bones.org[0])
        self.configure_control_bone(2, self.bones.ctrl.head, self.bones.org[-1])
        if self.long_neck:
            self.configure_neck_bend_bone(self.bones.ctrl.neck_bend, self.bones.org[0])

    def configure_neck_bend_bone(self, ctrl: str, _org: str):
        bone = self.get_bone(ctrl)
        bone.lock_rotation = (True, True, True)
        bone.lock_rotation_w = True
        bone.lock_scale = (True, True, True)

    @stage.generate_widgets
    def make_control_widgets(self):
        ctrl = self.bones.ctrl
        if self.has_neck:
            self.make_neck_widget(ctrl.neck)
        self.make_head_widget(ctrl.head)
        if self.long_neck:
            self.make_neck_bend_widget(ctrl.neck_bend)

    def make_neck_widget(self, ctrl: str):
        radius = 1 / max(1, len(self.bones.mch.chain))

        create_circle_widget(
            self.obj, ctrl,
            radius=radius,
            head_tail=0.5,
        )

    def make_neck_bend_widget(self, ctrl: str):
        radius = 1 / max(1, len(self.bones.mch.chain))

        create_neck_bend_widget(
            self.obj, ctrl,
            radius=radius / 2,
            head_tail=0.0,
        )

    def make_head_widget(self, ctrl: str):
        # place wgt @ middle of head bone for long necks
        if self.long_neck:
            head_tail = 0.5
        else:
            head_tail = 1.0

        create_circle_widget(
            self.obj, ctrl,
            radius=0.5,
            head_tail=head_tail,
            with_line=False,
        )

    ####################################################
    # MCH bones associated with main controls

    @stage.generate_bones
    def make_mch_control_bones(self):
        orgs = self.bones.org
        mch = self.bones.mch

        if self.has_neck:
            mch.rot_neck = self.make_mch_follow_bone(orgs[0], 'neck', 0.5, copy_scale=True)
            mch.stretch = self.make_mch_stretch_bone(orgs[0], 'STR-neck', orgs[-1])
        mch.rot_head = self.make_mch_follow_bone(orgs[-1], 'head', 0.0, copy_scale=True)

    def make_mch_stretch_bone(self, org: str, name: str, org_head: str):
        name = self.copy_bone(org, make_derived_name(name, 'mch'), parent=False)
        self.get_bone(name).tail = self.get_bone(org_head).head
        return name

    @stage.parent_bones
    def parent_mch_control_bones(self):
        if self.has_neck:
            self.set_bone_parent(self.bones.mch.rot_neck, self.rig_parent_bone)
            self.set_bone_parent(self.bones.mch.rot_head, self.bones.ctrl.neck)
            self.set_bone_parent(self.bones.mch.stretch, self.bones.ctrl.neck)
        else:
            self.set_bone_parent(self.bones.mch.rot_head, self.rig_parent_bone)

    @stage.rig_bones
    def rig_mch_control_bones(self):
        if self.has_neck:
            self.rig_mch_stretch_bone(self.bones.mch.stretch, self.bones.ctrl.head)

    def rig_mch_stretch_bone(self, mch: str, head: str):
        self.make_constraint(mch, 'STRETCH_TO', head, keep_axis='SWING_Y')

    ####################################################
    # MCH IK chain for the long neck

    @stage.generate_bones
    def make_mch_ik_chain(self):
        orgs = self.bones.org
        if self.long_neck:
            self.bones.mch.ik = map_list(self.make_mch_ik_bone, orgs[0:-1])

    def make_mch_ik_bone(self, org: str):
        return self.copy_bone(org, make_derived_name(org, 'mch', '_ik'), parent=False)

    @stage.parent_bones
    def parent_mch_ik_chain(self):
        if self.long_neck:
            ik = self.bones.mch.ik
            self.set_bone_parent(ik[0], self.bones.ctrl.tweak[0])
            self.parent_bone_chain(ik, use_connect=True)

    @stage.rig_bones
    def rig_mch_ik_chain(self):
        if self.long_neck:
            ik = self.bones.mch.ik
            head = self.bones.ctrl.head
            for args in zip(count(0), ik):
                self.rig_mch_ik_bone(*args, len(ik), head)

    def rig_mch_ik_bone(self, i: int, mch: str, ik_len: int, head: str):
        if i == ik_len - 1:
            self.make_constraint(mch, 'IK', head, chain_count=ik_len)

        self.get_bone(mch).ik_stretch = 0.1

    ####################################################
    # MCH chain for the middle of the neck

    @stage.generate_bones
    def make_mch_chain(self):
        orgs = self.bones.org
        self.bones.mch.chain = map_list(self.make_mch_bone, orgs[1:-1])

    def make_mch_bone(self, org: str):
        return self.copy_bone(org, make_derived_name(org, 'mch'), parent=False, scale=1 / 4)

    @stage.parent_bones
    def align_mch_chain(self):
        for mch in self.bones.mch.chain:
            align_bone_orientation(self.obj, mch, self.bones.ctrl.neck)

    @stage.parent_bones
    def parent_mch_chain(self):
        mch = self.bones.mch
        for bone in mch.chain:
            self.set_bone_parent(bone, mch.stretch, inherit_scale='NONE')

    @stage.rig_bones
    def rig_mch_chain(self):
        chain = self.bones.mch.chain
        if self.long_neck:
            ik = self.bones.mch.ik
            for args in zip(count(0), chain, ik[1:]):
                self.rig_mch_bone_long(*args, len(chain))
        else:
            for args in zip(count(0), chain):
                self.rig_mch_bone(*args, len(chain))

    def rig_mch_bone_long(self, i: int, mch: str, ik: str, len_mch: int):
        ctrl = self.bones.ctrl

        self.make_constraint(mch, 'COPY_LOCATION', ik)

        step = 2 / (len_mch + 1)
        x_val = (i + 1) * step
        influence = 2 * x_val - x_val**2    # parabolic influence of pivot

        self.make_constraint(
            mch, 'COPY_LOCATION', ctrl.neck_bend,
            influence=influence, use_offset=True, space='LOCAL'
        )

        self.make_constraint(mch, 'COPY_SCALE', ctrl.neck)

    def rig_mch_bone(self, i, mch, len_mch):
        ctrl = self.bones.ctrl

        n_factor = float((i + 1) / (len_mch + 1))
        self.make_constraint(
            mch, 'COPY_ROTATION', ctrl.head,
            influence=n_factor, space='LOCAL'
        )

        self.make_constraint(mch, 'COPY_SCALE', ctrl.neck)

    ####################################################
    # Tweak bones

    @stage.generate_bones
    def make_tweak_chain(self):
        orgs = self.bones.org
        self.bones.ctrl.tweak = map_list(self.make_tweak_bone, count(0), orgs[0:-1])
        if not self.has_neck:
            self.check_connect_tweak(orgs[0])

    @stage.parent_bones
    def parent_tweak_chain(self):
        ctrl = self.bones.ctrl
        mch = self.bones.mch

        if self.has_neck:
            for args in zip(ctrl.tweak, [ctrl.neck, *mch.chain]):
                self.set_bone_parent(*args)
        elif self.connected_tweak:
            self.set_bone_parent(self.connected_tweak, ctrl.head)

    @stage.rig_bones
    def generate_neck_tweak_widget(self):
        # Generate the widget early to override connected parent
        if self.long_neck:
            bone = self.bones.ctrl.tweak[0]
            create_neck_tweak_widget(self.obj, bone, size=1.0)

    ####################################################
    # ORG and DEF bones

    @stage.generate_bones
    def register_parent_bones(self):
        rig = self.rigify_parent or self
        builder = SwitchParentBuilder(self.generator)
        builder.register_parent(
            self, self.bones.org[-1], name='Head',
            inject_into=rig, exclude_self=True, tags={'head'},
        )

    @stage.configure_bones
    def configure_bbone_chain(self):
        self.get_bone(self.bones.deform[-1]).bone.bbone_segments = 1

    @stage.rig_bones
    def rig_org_chain(self):
        if self.has_neck:
            tweaks = self.bones.ctrl.tweak + [self.bones.ctrl.head]
        else:
            tweaks = [self.connected_tweak or self.bones.ctrl.head]

        for args in zip(count(0), self.bones.org, tweaks, [*tweaks[1:], None]):
            self.rig_org_bone(*args)


def create_sample(obj, *, parent=None):
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('neck')
    bone.head[:] = 0.0000, 0.0114, 1.6582
    bone.tail[:] = 0.0000, -0.0130, 1.7197
    bone.roll = 0.0000
    bone.use_connect = False
    if parent:
        bone.parent = arm.edit_bones[parent]
    bones['neck'] = bone.name
    bone = arm.edit_bones.new('neck.001')
    bone.head[:] = 0.0000, -0.0130, 1.7197
    bone.tail[:] = 0.0000, -0.0247, 1.7813
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['neck']]
    bones['neck.001'] = bone.name
    bone = arm.edit_bones.new('head')
    bone.head[:] = 0.0000, -0.0247, 1.7813
    bone.tail[:] = 0.0000, -0.0247, 1.9796
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['neck.001']]
    bones['head'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['neck']]
    pbone.rigify_type = 'spines.super_head'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.connect_chain = bool(parent)
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['neck.001']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['head']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone
        if bcoll := arm.collections.active:
            bcoll.assign(bone)



# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from bpy.types import (
    Panel,
)
from bl_ui.properties_physics_common import (
    point_cache_ui,
    effector_weights_ui,
)

COMPAT_OB_TYPES = {'MESH', 'LATTICE', 'CURVE', 'SURFACE', 'FONT'}


def softbody_panel_enabled(md):
    return (md.point_cache.is_baked is False)


class PhysicButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "physics"

    @classmethod
    def poll(cls, context):
        ob = context.object
        return ob and ob.type in COMPAT_OB_TYPES and context.engine in cls.COMPAT_ENGINES and context.soft_body


class PHYSICS_PT_softbody(PhysicButtonsPanel, Panel):
    bl_label = "Soft Body"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.prop(softbody, "collision_collection")


class PHYSICS_PT_softbody_object(PhysicButtonsPanel, Panel):
    bl_label = "Object"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings
        ob = context.object

        layout.enabled = softbody_panel_enabled(md)
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column()
        col.prop(softbody, "friction")

        col.separator()

        col = flow.column()
        col.prop(softbody, "mass")

        col.prop_search(softbody, "vertex_group_mass", ob, "vertex_groups", text="Control Point")


class PHYSICS_PT_softbody_simulation(PhysicButtonsPanel, Panel):
    bl_label = "Simulation"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.enabled = softbody_panel_enabled(md)

        layout.prop(softbody, "speed")


class PHYSICS_PT_softbody_cache(PhysicButtonsPanel, Panel):
    bl_label = "Cache"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        md = context.soft_body
        point_cache_ui(self, md.point_cache, softbody_panel_enabled(md), 'SOFTBODY')


class PHYSICS_PT_softbody_goal(PhysicButtonsPanel, Panel):
    bl_label = "Goal"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        softbody = context.soft_body.settings

        self.layout.active = softbody_panel_enabled(context.soft_body)
        self.layout.prop(softbody, "use_goal", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings
        ob = context.object

        layout.active = softbody.use_goal and softbody_panel_enabled(md)

        layout.prop_search(softbody, "vertex_group_goal", ob, "vertex_groups", text="Vertex Group")


class PHYSICS_PT_softbody_goal_strengths(PhysicButtonsPanel, Panel):
    bl_label = "Strengths"
    bl_parent_id = "PHYSICS_PT_softbody_goal"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody.use_goal and softbody_panel_enabled(md)
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column()
        col.prop(softbody, "goal_default", text="Default")

        col.separator()

        col = flow.column(align=True)
        col.prop(softbody, "goal_min", text="Min")
        col.prop(softbody, "goal_max", text="Max")


class PHYSICS_PT_softbody_goal_settings(PhysicButtonsPanel, Panel):
    bl_label = "Settings"
    bl_parent_id = "PHYSICS_PT_softbody_goal"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody.use_goal and softbody_panel_enabled(md)
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column()
        col.prop(softbody, "goal_spring", text="Stiffness")

        col = flow.column()
        col.prop(softbody, "goal_friction", text="Damping")


class PHYSICS_PT_softbody_edge(PhysicButtonsPanel, Panel):
    bl_label = "Edges"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        softbody = context.soft_body.settings

        self.layout.active = softbody_panel_enabled(context.soft_body)
        self.layout.prop(softbody, "use_edges", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings
        ob = context.object

        layout.active = softbody.use_edges and softbody_panel_enabled(md)
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column()

        col.prop_search(softbody, "vertex_group_spring", ob, "vertex_groups", text="Springs")

        col.separator()

        col.prop(softbody, "pull")
        col.prop(softbody, "push")

        col.separator()

        col = flow.column()
        col.prop(softbody, "damping")
        col.prop(softbody, "plastic")
        col.prop(softbody, "bend")

        col.separator()

        col = flow.column()
        col.prop(softbody, "spring_length", text="Length")

        col.separator()

        col = flow.column(align=True, heading="Collision")
        col.prop(softbody, "use_edge_collision", text="Edge", toggle=False)
        col.prop(softbody, "use_face_collision", text="Face", toggle=False)


class PHYSICS_PT_softbody_edge_aerodynamics(PhysicButtonsPanel, Panel):
    bl_label = "Aerodynamics"
    bl_parent_id = "PHYSICS_PT_softbody_edge"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        md = context.soft_body
        softbody = md.settings

        flow.active = softbody.use_edges and softbody_panel_enabled(md)

        col = flow.column()
        col.prop(softbody, "aerodynamics_type", text="Type")

        col = flow.column()
        col.prop(softbody, "aero", text="Factor")


class PHYSICS_PT_softbody_edge_stiffness(PhysicButtonsPanel, Panel):
    bl_label = "Stiffness"
    bl_parent_id = "PHYSICS_PT_softbody_edge"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        softbody = context.soft_body.settings

        self.layout.active = softbody_panel_enabled(context.soft_body)
        self.layout.prop(softbody, "use_stiff_quads", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody.use_edges and softbody.use_stiff_quads and softbody_panel_enabled(md)

        layout.prop(softbody, "shear")


class PHYSICS_PT_softbody_collision(PhysicButtonsPanel, Panel):
    bl_label = "Self Collision"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):
        softbody = context.soft_body.settings

        self.layout.active = softbody_panel_enabled(context.soft_body)
        self.layout.prop(softbody, "use_self_collision", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody.use_self_collision and softbody_panel_enabled(md)

        layout.prop(softbody, "collision_type", text="Calculation Type")

        layout.separator()

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column()
        col.prop(softbody, "ball_size", text="Ball Size")

        col = flow.column()
        col.prop(softbody, "ball_stiff", text="Stiffness")
        col.prop(softbody, "ball_damp", text="Dampening")


class PHYSICS_PT_softbody_solver(PhysicButtonsPanel, Panel):
    bl_label = "Solver"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody_panel_enabled(md)
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column(align=True)
        col.prop(softbody, "step_min", text="Step Size Min")
        col.prop(softbody, "step_max", text="Max")

        col = flow.column()
        col.prop(softbody, "use_auto_step", text="Auto-Step")
        col.prop(softbody, "error_threshold")


class PHYSICS_PT_softbody_solver_diagnostics(PhysicButtonsPanel, Panel):
    bl_label = "Diagnostics"
    bl_parent_id = "PHYSICS_PT_softbody_solver"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody_panel_enabled(md)

        layout.prop(softbody, "use_diagnose")
        layout.prop(softbody, "use_estimate_matrix")


class PHYSICS_PT_softbody_solver_helpers(PhysicButtonsPanel, Panel):
    bl_label = "Helpers"
    bl_parent_id = "PHYSICS_PT_softbody_solver"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.soft_body
        softbody = md.settings

        layout.active = softbody_panel_enabled(md)
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        col = flow.column()
        col.prop(softbody, "choke")

        col = flow.column()
        col.prop(softbody, "fuzzy")


class PHYSICS_PT_softbody_field_weights(PhysicButtonsPanel, Panel):
    bl_label = "Field Weights"
    bl_parent_id = "PHYSICS_PT_softbody"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        md = context.soft_body
        softbody = md.settings

        effector_weights_ui(self, softbody.effector_weights, 'SOFTBODY')


classes = (
    PHYSICS_PT_softbody,
    PHYSICS_PT_softbody_object,
    PHYSICS_PT_softbody_simulation,
    PHYSICS_PT_softbody_cache,
    PHYSICS_PT_softbody_goal,
    PHYSICS_PT_softbody_goal_settings,
    PHYSICS_PT_softbody_goal_strengths,
    PHYSICS_PT_softbody_edge,
    PHYSICS_PT_softbody_edge_aerodynamics,
    PHYSICS_PT_softbody_edge_stiffness,
    PHYSICS_PT_softbody_collision,
    PHYSICS_PT_softbody_solver,
    PHYSICS_PT_softbody_solver_diagnostics,
    PHYSICS_PT_softbody_solver_helpers,
    PHYSICS_PT_softbody_field_weights,
)


if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



# SPDX-FileCopyrightText: 2018-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from bpy.types import (
    Operator,
)
from bpy.props import (
    IntProperty,
)


class CONSTRAINT_OT_add_target(Operator):
    """Add a target to the constraint"""
    bl_idname = "constraint.add_target"
    bl_label = "Add Target"
    bl_options = {'UNDO', 'INTERNAL'}

    @classmethod
    def poll(cls, context):
        constraint = getattr(context, "constraint", None)
        return constraint

    def execute(self, context):
        context.constraint.targets.new()
        return {'FINISHED'}


class CONSTRAINT_OT_remove_target(Operator):
    """Remove the target from the constraint"""
    bl_idname = "constraint.remove_target"
    bl_label = "Remove Target"
    bl_options = {'UNDO', 'INTERNAL'}

    index: IntProperty()

    @classmethod
    def poll(cls, context):
        constraint = getattr(context, "constraint", None)
        return constraint

    def execute(self, context):
        tgts = context.constraint.targets
        tgts.remove(tgts[self.index])
        return {'FINISHED'}


class CONSTRAINT_OT_normalize_target_weights(Operator):
    """Normalize weights of all target bones"""
    bl_idname = "constraint.normalize_target_weights"
    bl_label = "Normalize Weights"
    bl_options = {'UNDO', 'INTERNAL'}

    @classmethod
    def poll(cls, context):
        constraint = getattr(context, "constraint", None)
        return constraint

    def execute(self, context):
        tgts = context.constraint.targets
        total = sum(t.weight for t in tgts)

        if total > 0:
            for t in tgts:
                t.weight = t.weight / total

        return {'FINISHED'}


class CONSTRAINT_OT_disable_keep_transform(Operator):
    """Set the influence of this constraint to zero while """ \
        """trying to maintain the object's transformation. Other active """ \
        """constraints can still influence the final transformation"""

    bl_idname = "constraint.disable_keep_transform"
    bl_label = "Disable and Keep Transform"
    bl_options = {'UNDO', 'INTERNAL'}

    @classmethod
    def poll(cls, context):
        constraint = getattr(context, "constraint", None)
        return constraint and constraint.influence > 0.0

    def execute(self, context):
        """Disable constraint while maintaining the visual transform"""

        # This works most of the time, but when there are multiple constraints active
        # there could still be one that overrides the visual transform.
        #
        # Note that executing this operator and then increasing the constraint
        # influence may move the object; this happens when the constraint is
        # additive rather than replacing the transform entirely.

        # Get the matrix in world space.
        is_bone_constraint = context.space_data.context == 'BONE_CONSTRAINT'
        ob = context.object
        if is_bone_constraint:
            bone = context.pose_bone
            mat = ob.matrix_world @ bone.matrix
        else:
            mat = ob.matrix_world

        context.constraint.influence = 0.0

        # Set the matrix.
        if is_bone_constraint:
            bone.matrix = ob.matrix_world.inverted() @ mat
        else:
            ob.matrix_world = mat

        return {'FINISHED'}


classes = (
    CONSTRAINT_OT_add_target,
    CONSTRAINT_OT_remove_target,
    CONSTRAINT_OT_normalize_target_weights,
    CONSTRAINT_OT_disable_keep_transform,
)



# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import (
    Header,
    Menu,
    Panel,
)

from bl_ui.properties_data_grease_pencil import (
    GreasePencil_LayerMaskPanel,
    GreasePencil_LayerTransformPanel,
    GreasePencil_LayerRelationsPanel,
    GreasePencil_LayerAdjustmentsPanel,
    GreasePencil_LayerDisplayPanel,
)
from bl_ui.space_time import playback_controls
from bl_ui.properties_data_mesh import draw_shape_key_properties

from rna_prop_ui import PropertyPanel

#######################################
# DopeSheet Filtering - Header Buttons

# used for DopeSheet, NLA, and Graph Editors


def dopesheet_filter(layout, context):
    dopesheet = context.space_data.dopesheet
    is_nla = context.area.type == 'NLA_EDITOR'
    is_action_editor = not is_nla and context.space_data.mode == 'ACTION'

    row = layout.row(align=True)
    if is_action_editor:
        row.prop(dopesheet, "show_only_slot_of_active_object", text="")
    row.prop(dopesheet, "show_only_selected", text="")
    row.prop(dopesheet, "show_hidden", text="")

    if is_nla:
        row.prop(dopesheet, "show_missing_nla", text="")
    else:  # graph and dopesheet editors - F-Curves and drivers only
        row.prop(dopesheet, "show_only_errors", text="")

#######################################
# Dope-sheet Filtering Popovers

# Generic Layout - Used as base for filtering popovers used in all animation editors
# Used for DopeSheet, NLA, and Graph Editors


class DopesheetFilterPopoverBase:
    bl_region_type = 'HEADER'
    bl_label = "Filters"

    # Generic = Affects all data-types.
    # XXX: Perhaps we want these to stay in the header instead, for easy/fast access
    @classmethod
    def draw_generic_filters(cls, context, layout):
        dopesheet = context.space_data.dopesheet
        is_nla = context.area.type == 'NLA_EDITOR'

        col = layout.column(align=True)
        col.prop(dopesheet, "show_only_selected", icon='NONE')
        col.prop(dopesheet, "show_hidden", icon='NONE')

        if is_nla:
            col.prop(dopesheet, "show_missing_nla", icon='NONE')
        else:  # Graph and dope-sheet editors - F-Curves and drivers only.
            col.prop(dopesheet, "show_only_errors", icon='NONE')

    # Name/Membership Filters
    # XXX: Perhaps these should just stay in the headers (exclusively)?
    @classmethod
    def draw_search_filters(cls, context, layout, generic_filters_only=False):
        dopesheet = context.space_data.dopesheet
        is_nla = context.area.type == 'NLA_EDITOR'

        col = layout.column(align=True)
        if not is_nla:
            row = col.row(align=True)
            row.prop(dopesheet, "filter_fcurve_name", text="")
        else:
            row = col.row(align=True)
            row.prop(dopesheet, "filter_text", text="")

        if (not generic_filters_only) and bpy.data.collections:
            col = layout.column(align=True)
            col.prop(dopesheet, "filter_collection", text="")

    # Standard = Present in all panels
    @classmethod
    def draw_standard_filters(cls, context, layout):
        dopesheet = context.space_data.dopesheet

        # datablock filters
        layout.label(text="Filter by Type:")
        flow = layout.grid_flow(row_major=True, columns=2, even_rows=False, align=False)

        flow.prop(dopesheet, "show_scenes", text="Scenes")
        flow.prop(dopesheet, "show_nodes", text="Node Trees")

        # object types
        if bpy.data.armatures:
            flow.prop(dopesheet, "show_armatures", text="Armatures")
        if bpy.data.cameras:
            flow.prop(dopesheet, "show_cameras", text="Cameras")
        if bpy.data.grease_pencils:
            flow.prop(dopesheet, "show_gpencil", text="Grease Pencil Objects")
        if bpy.data.lights:
            flow.prop(dopesheet, "show_lights", text="Lights")
        if bpy.data.meshes:
            flow.prop(dopesheet, "show_meshes", text="Meshes")
        if bpy.data.curves:
            flow.prop(dopesheet, "show_curves", text="Curves")
        if bpy.data.lattices:
            flow.prop(dopesheet, "show_lattices", text="Lattices")
        if bpy.data.metaballs:
            flow.prop(dopesheet, "show_metaballs", text="Metaballs")
        if hasattr(bpy.data, "hair_curves") and bpy.data.hair_curves:
            flow.prop(dopesheet, "show_hair_curves", text="Hair Curves")
        if hasattr(bpy.data, "pointclouds") and bpy.data.pointclouds:
            flow.prop(dopesheet, "show_pointclouds", text="Point Clouds")
        if bpy.data.volumes:
            flow.prop(dopesheet, "show_volumes", text="Volumes")
        if bpy.data.lightprobes:
            flow.prop(dopesheet, "show_lightprobes", text="Light Probes")

        # data types
        flow.prop(dopesheet, "show_worlds", text="Worlds")
        if bpy.data.particles:
            flow.prop(dopesheet, "show_particles", text="Particles")
        if bpy.data.linestyles:
            flow.prop(dopesheet, "show_linestyles", text="Line Styles")
        if bpy.data.speakers:
            flow.prop(dopesheet, "show_speakers", text="Speakers")
        if bpy.data.materials:
            flow.prop(dopesheet, "show_materials", text="Materials")
        if bpy.data.textures:
            flow.prop(dopesheet, "show_textures", text="Textures")
        if bpy.data.shape_keys:
            flow.prop(dopesheet, "show_shapekeys", text="Shape Keys")
        if bpy.data.cache_files:
            flow.prop(dopesheet, "show_cache_files", text="Cache Files")
        if bpy.data.movieclips:
            flow.prop(dopesheet, "show_movieclips", text="Movie Clips")

        layout.separator()

        # Object Data Filters

        # TODO: Add per-channel/axis convenience toggles?
        split = layout.split()

        col = split.column()
        col.prop(dopesheet, "show_transforms", text="Transforms")

        col = split.column()
        col.prop(dopesheet, "show_modifiers", text="Modifiers")

        layout.separator()

        # performance-related options (users will mostly have these enabled)
        col = layout.column(align=True)
        col.label(text="Options:")
        col.prop(dopesheet, "use_datablock_sort", icon='NONE')


# Popover for Dope-sheet Editor(s) - Dope-sheet, Action, Shape-key, GPencil, Mask, etc.
class DOPESHEET_PT_filters(DopesheetFilterPopoverBase, Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'HEADER'
    bl_label = "Filters"

    def draw(self, context):
        layout = self.layout

        dopesheet = context.space_data.dopesheet
        ds_mode = context.space_data.mode

        layout.prop(dopesheet, "show_summary", text="Summary")

        DopesheetFilterPopoverBase.draw_generic_filters(context, layout)

        if ds_mode in {'DOPESHEET', 'ACTION', 'GPENCIL'}:
            layout.separator()
            generic_filters_only = ds_mode != 'DOPESHEET'
            DopesheetFilterPopoverBase.draw_search_filters(context, layout, generic_filters_only=generic_filters_only)

        if ds_mode == 'DOPESHEET':
            layout.separator()
            DopesheetFilterPopoverBase.draw_standard_filters(context, layout)


#######################################
# DopeSheet Editor - General/Standard UI

class DOPESHEET_HT_header(Header):
    bl_space_type = 'DOPESHEET_EDITOR'

    def draw(self, context):
        layout = self.layout

        st = context.space_data

        layout.template_header()

        if st.mode != 'TIMELINE':
            # Timeline mode is special, as it's presented as a sub-type of the
            # dope sheet editor, rather than a mode. So this shouldn't show the
            # mode selector.
            layout.prop(st, "ui_mode", text="")

        DOPESHEET_MT_editor_menus.draw_collapsible(context, layout)
        DOPESHEET_HT_editor_buttons.draw_header(context, layout)


# Header for "normal" dopesheet editor modes (e.g. Dope Sheet, Action, Shape Keys, etc.)
class DOPESHEET_HT_editor_buttons:

    @classmethod
    def draw_header(cls, context, layout):
        st = context.space_data

        if st.mode == 'TIMELINE':
            playback_controls(layout, context)
            layout.separator()
            cls._draw_overlay_selector(context, layout)
            return

        if st.mode in {'ACTION', 'SHAPEKEY'} and context.object:
            layout.separator_spacer()
            cls._draw_action_selector(context, layout)

        # Layer management
        if st.mode == 'GPENCIL':
            ob = context.active_object

            enable_but = ob is not None and ob.type == 'GREASEPENCIL'

            row = layout.row(align=True)
            row.enabled = enable_but
            row.operator("grease_pencil.layer_add", icon='ADD', text="")
            row.operator("grease_pencil.layer_remove", icon='REMOVE', text="")
            row.menu("GREASE_PENCIL_MT_grease_pencil_add_layer_extra", icon='DOWNARROW_HLT', text="")

            row = layout.row(align=True)
            row.enabled = enable_but
            row.operator("anim.channels_move", icon='TRIA_UP', text="").direction = 'UP'
            row.operator("anim.channels_move", icon='TRIA_DOWN', text="").direction = 'DOWN'

            row = layout.row(align=True)
            row.enabled = enable_but
            row.operator("grease_pencil.layer_isolate", icon='RESTRICT_VIEW_ON', text="").affect_visibility = True
            row.operator("grease_pencil.layer_isolate", icon='LOCKED', text="").affect_visibility = False

        layout.separator_spacer()

        if st.mode == 'DOPESHEET':
            dopesheet_filter(layout, context)
        elif st.mode == 'ACTION':
            dopesheet_filter(layout, context)
        elif st.mode == 'GPENCIL':
            row = layout.row(align=True)
            row.prop(st.dopesheet, "show_only_selected", text="")
            row.prop(st.dopesheet, "show_hidden", text="")

        layout.popover(
            panel="DOPESHEET_PT_filters",
            text="",
            icon='FILTER',
        )

        tool_settings = context.tool_settings

        # Grease Pencil mode doesn't need snapping, as it's frame-aligned only
        if st.mode != 'GPENCIL':
            row = layout.row(align=True)
            row.prop(tool_settings, "use_snap_anim", text="")
            sub = row.row(align=True)
            sub.popover(
                panel="DOPESHEET_PT_snapping",
                text="",
            )

        row = layout.row(align=True)
        row.prop(tool_settings, "use_proportional_action", text="", icon_only=True)
        sub = row.row(align=True)
        sub.active = tool_settings.use_proportional_action
        sub.prop_with_popover(
            tool_settings,
            "proportional_edit_falloff",
            text="",
            icon_only=True,
            panel="DOPESHEET_PT_proportional_edit",
        )

        cls._draw_overlay_selector(context, layout)

    @classmethod
    def _draw_overlay_selector(cls, context, layout):
        st = context.space_data

        overlays = st.overlays
        row = layout.row(align=True)
        row.prop(overlays, "show_overlays", text="", icon='OVERLAY')
        sub = row.row(align=True)
        sub.popover(panel="DOPESHEET_PT_overlay", text="")
        sub.active = overlays.show_overlays

    @classmethod
    def _draw_action_selector(cls, context, layout):
        animated_id = cls._get_animated_id(context)
        if not animated_id:
            return

        row = layout.row()
        if animated_id.animation_data and animated_id.animation_data.use_tweak_mode:
            row.enabled = False

        row.template_action(animated_id, new="action.new", unlink="action.unlink")

        adt = animated_id and animated_id.animation_data
        if not adt or not adt.action or not adt.action.is_action_layered:
            return

        # Store the animated ID in the context, so that the new/unlink operators
        # have access to it.
        row.context_pointer_set("animated_id", animated_id)
        row.template_search(
            adt, "action_slot",
            adt, "action_suitable_slots",
            new="anim.slot_new_for_id",
            unlink="anim.slot_unassign_from_id",
        )

    @staticmethod
    def _get_animated_id(context):
        st = context.space_data
        match st.mode:
            case 'ACTION':
                return context.object
            case 'SHAPEKEY':
                return getattr(context.object.data, "shape_keys", None)
            case _:
                print("Dope Sheet mode '{:s}' not expected to have an Action selector".format(st.mode))
                return context.object


class DOPESHEET_HT_playback_controls(Header):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'FOOTER'

    def draw(self, context):
        layout = self.layout

        playback_controls(layout, context)


class DOPESHEET_PT_snapping(Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'HEADER'
    bl_label = "Snapping"

    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.label(text="Snap To")
        tool_settings = context.tool_settings
        col.prop(tool_settings, "snap_anim_element", expand=True)
        if tool_settings.snap_anim_element != 'MARKER':
            col.prop(tool_settings, "use_snap_time_absolute")


class DOPESHEET_PT_proportional_edit(Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'HEADER'
    bl_label = "Proportional Editing"
    bl_ui_units_x = 8

    def draw(self, context):
        layout = self.layout
        tool_settings = context.tool_settings
        col = layout.column()
        col.active = tool_settings.use_proportional_action

        col.prop(tool_settings, "proportional_edit_falloff", expand=True)
        col.prop(tool_settings, "proportional_size")


class DOPESHEET_MT_editor_menus(Menu):
    bl_idname = "DOPESHEET_MT_editor_menus"
    bl_label = ""

    def draw(self, context):
        layout = self.layout
        st = context.space_data
        active_action = context.active_action

        if st.mode == 'TIMELINE':
            # Draw the 'timeline' menus, which are simpler. Most importantly, the
            # 'selected only' toggle is in the menu, and actually stored as a scene
            # flag instead of the space data.
            horizontal = (layout.direction == 'VERTICAL')
            if horizontal:
                row = layout.row()
                sub = row.row(align=True)
            else:
                sub = layout
            sub.menu("TIME_MT_view")
            if st.show_markers:
                sub.menu("DOPESHEET_MT_marker")
            return

        layout.menu("DOPESHEET_MT_view")
        layout.menu("DOPESHEET_MT_select")
        if st.show_markers:
            layout.menu("DOPESHEET_MT_marker")

        if st.mode == 'DOPESHEET' or (st.mode == 'ACTION' and active_action is not None):
            layout.menu("DOPESHEET_MT_channel")
        elif st.mode == 'GPENCIL':
            layout.menu("DOPESHEET_MT_gpencil_channel")

        layout.menu("DOPESHEET_MT_key")

        if st.mode in {'ACTION', 'SHAPEKEY'} and active_action is not None:
            layout.menu("DOPESHEET_MT_action")


class DOPESHEET_MT_view(Menu):
    bl_label = "View"

    def draw(self, context):
        layout = self.layout

        st = context.space_data

        layout.prop(st, "show_region_ui")
        layout.prop(st, "show_region_hud")
        layout.prop(st, "show_region_channels")
        layout.prop(st, "show_region_footer", text="Playback Controls")
        layout.separator()

        layout.operator("action.view_selected")
        layout.operator("action.view_all")
        if context.scene.use_preview_range:
            layout.operator("anim.scene_range_frame", text="Frame Preview Range")
        else:
            layout.operator("anim.scene_range_frame", text="Frame Scene Range")
        layout.operator("action.view_frame")
        layout.separator()

        layout.prop(st.dopesheet, "use_multi_word_filter", text="Multi-Word Match Search")
        layout.separator()

        layout.prop(st, "use_realtime_update")

        # Sliders are always shown in the Shape Key Editor regardless of this setting.
        col = layout.column()
        col.active = context.space_data.mode != 'SHAPEKEY'
        col.prop(st, "show_sliders")

        layout.prop(st, "show_interpolation")
        layout.prop(st, "show_extremes")
        layout.prop(st, "use_auto_merge_keyframes")
        layout.separator()

        layout.prop(st, "show_markers")
        layout.prop(st, "show_seconds")
        layout.prop(st, "show_locked_time")
        layout.separator()

        layout.operator("anim.previewrange_set")
        layout.operator("anim.previewrange_clear")
        layout.operator("action.previewrange_set")
        layout.separator()

        # Add this to show key-binding (reverse action in dope-sheet).
        props = layout.operator("wm.context_set_enum", text="Toggle Graph Editor", icon='GRAPH')
        props.data_path = "area.type"
        props.value = 'GRAPH_EDITOR'
        layout.separator()

        layout.menu("DOPESHEET_MT_cache")
        layout.separator()

        layout.menu("INFO_MT_area")


class DOPESHEET_MT_view_pie(Menu):
    bl_label = "View"

    def draw(self, context):
        layout = self.layout

        pie = layout.menu_pie()
        pie.operator("action.view_all")
        pie.operator("action.view_selected", icon='ZOOM_SELECTED')
        pie.operator("action.view_frame")
        if context.scene.use_preview_range:
            pie.operator("anim.scene_range_frame", text="Frame Preview Range")
        else:
            pie.operator("anim.scene_range_frame", text="Frame Scene Range")


class DOPESHEET_MT_cache(Menu):
    bl_label = "Cache"

    def draw(self, context):
        layout = self.layout

        st = context.space_data

        layout.prop(st, "show_cache")

        layout.separator()

        col = layout.column()
        col.enabled = st.show_cache
        col.prop(st, "cache_softbody")
        col.prop(st, "cache_particles")
        col.prop(st, "cache_cloth")
        col.prop(st, "cache_simulation_nodes")
        col.prop(st, "cache_smoke")
        col.prop(st, "cache_dynamicpaint")
        col.prop(st, "cache_rigidbody")


class DOPESHEET_MT_select(Menu):
    bl_label = "Select"

    def draw(self, context):
        layout = self.layout

        layout.operator("action.select_all", text="All").action = 'SELECT'
        layout.operator("action.select_all", text="None").action = 'DESELECT'
        layout.operator("action.select_all", text="Invert").action = 'INVERT'

        layout.separator()
        layout.operator("action.select_box").axis_range = False
        layout.operator("action.select_box", text="Box Select (Axis Range)").axis_range = True
        layout.operator("action.select_circle")
        layout.operator_menu_enum("action.select_lasso", "mode")

        # FIXME: grease pencil mode isn't supported for these yet, so skip for that mode only
        if context.space_data.mode != 'GPENCIL':
            layout.separator()
            layout.operator("action.select_more", text="More")
            layout.operator("action.select_less", text="Less")

            layout.separator()
            layout.operator("action.select_linked")

        layout.separator()
        layout.operator("action.select_column", text="Columns on Selected Keys").mode = 'KEYS'
        layout.operator("action.select_column", text="Column on Current Frame").mode = 'CFRA'

        layout.operator("action.select_column", text="Columns on Selected Markers").mode = 'MARKERS_COLUMN'
        layout.operator("action.select_column", text="Between Selected Markers").mode = 'MARKERS_BETWEEN'

        layout.separator()
        props = layout.operator("action.select_leftright", text="Before Current Frame")
        props.extend = False
        props.mode = 'LEFT'
        props = layout.operator("action.select_leftright", text="After Current Frame")
        props.extend = False
        props.mode = 'RIGHT'


class DOPESHEET_MT_marker(Menu):
    bl_label = "Marker"

    def draw(self, context):
        layout = self.layout

        from bl_ui.space_time import marker_menu_generic
        marker_menu_generic(layout, context)

        st = context.space_data

        if st.mode in {'ACTION', 'SHAPEKEY'} and context.active_action:
            layout.separator()
            layout.prop(st, "show_pose_markers")

            if st.show_pose_markers is False:
                layout.operator("action.markers_make_local")

        layout.prop(st, "use_marker_sync")

#######################################
# Keyframe Editing


class DOPESHEET_MT_channel(Menu):
    bl_label = "Channel"

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = 'INVOKE_REGION_CHANNELS'

        layout.operator("anim.channels_delete")
        layout.operator("action.clean", text="Clean Channels").channels = True

        layout.separator()
        layout.operator("anim.channels_group")
        layout.operator("anim.channels_ungroup")

        layout.separator()
        layout.operator_menu_enum("anim.channels_setting_toggle", "type")
        layout.operator_menu_enum("anim.channels_setting_enable", "type")
        layout.operator_menu_enum("anim.channels_setting_disable", "type")

        layout.separator()
        layout.operator("anim.channels_editable_toggle")
        layout.operator_menu_enum("action.extrapolation_type", "type", text="Extrapolation Mode")

        layout.separator()
        layout.operator("anim.channels_expand")
        layout.operator("anim.channels_collapse")

        layout.separator()
        layout.operator_menu_enum("anim.channels_move", "direction", text="Move...")

        layout.separator()
        layout.operator("anim.channels_fcurves_enable")

        layout.separator()
        layout.operator("anim.channels_bake")

        layout.separator()
        layout.operator("anim.channels_view_selected")


class DOPESHEET_MT_action(Menu):
    bl_label = "Action"

    def draw(self, context):
        layout = self.layout
        layout.operator("anim.merge_animation")
        layout.operator("anim.separate_slots")

        layout.separator()
        layout.operator("anim.slot_channels_move_to_new_action")

        layout.separator()
        layout.operator("action.push_down", text="Push Down Action", icon='NLA_PUSHDOWN')
        layout.operator("action.stash", text="Stash Action", icon='FREEZE')


class DOPESHEET_MT_key(Menu):
    bl_label = "Key"

    def draw(self, context):
        layout = self.layout
        ob = context.active_object

        layout.menu("DOPESHEET_MT_key_transform", text="Transform")

        layout.operator_menu_enum("action.snap", "type", text="Snap")
        layout.operator_menu_enum("action.mirror", "type", text="Mirror")

        layout.separator()
        layout.operator("action.keyframe_insert")

        layout.separator()
        layout.operator("action.frame_jump")

        layout.separator()
        layout.operator("action.copy")
        layout.operator("action.paste")
        layout.operator("action.paste", text="Paste Flipped").flipped = True
        layout.operator("action.duplicate_move")
        layout.operator("action.delete")
        if ob and ob.type == 'GREASEPENCIL':
            layout.operator("grease_pencil.delete_breakdown")

        layout.separator()
        layout.operator_menu_enum("action.keyframe_type", "type", text="Keyframe Type")
        layout.operator_menu_enum("action.handle_type", "type", text="Handle Type")
        layout.operator_menu_enum("action.interpolation_type", "type", text="Interpolation Mode")
        layout.operator_menu_enum("action.easing_type", "type", text="Easing Mode")

        layout.separator()
        layout.operator("action.clean").channels = False
        layout.operator("action.bake_keys")

        layout.separator()
        layout.operator("graph.euler_filter", text="Discontinuity (Euler) Filter")


class DOPESHEET_MT_key_transform(Menu):
    bl_label = "Transform"

    def draw(self, _context):
        layout = self.layout

        layout.operator("transform.transform", text="Move").mode = 'TIME_TRANSLATE'
        layout.operator("transform.transform", text="Extend").mode = 'TIME_EXTEND'
        layout.operator("transform.transform", text="Slide").mode = 'TIME_SLIDE'
        layout.operator("transform.transform", text="Scale").mode = 'TIME_SCALE'


class DopesheetActionPanelBase:
    bl_region_type = 'UI'
    bl_label = "Action"

    @classmethod
    def draw_generic_panel(cls, _context, layout, action):
        layout.label(text=action.name, icon='ACTION', translate=False)

        layout.prop(action, "use_frame_range")

        col = layout.column()
        col.active = action.use_frame_range

        row = col.row(align=True)
        row.prop(action, "frame_start", text="Start")
        row.prop(action, "frame_end", text="End")

        col.prop(action, "use_cyclic")


class DOPESHEET_PT_custom_props_action(PropertyPanel, Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_category = "Action"
    bl_region_type = 'UI'
    bl_context = "data"
    _context_path = "active_action"
    _property_type = bpy.types.Action

    @classmethod
    def poll(cls, context):
        return bool(context.active_action)


class DOPESHEET_PT_action(DopesheetActionPanelBase, Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_category = "Action"

    @classmethod
    def poll(cls, context):
        return bool(context.active_action)

    def draw(self, context):
        action = context.active_action
        self.draw_generic_panel(context, self.layout, action)


class DOPESHEET_PT_action_slot(Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'UI'
    bl_category = "Action"
    bl_label = "Slot"

    @classmethod
    def poll(cls, context):
        action = context.active_action
        return bool(action and action.slots.active)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        action = context.active_action
        slot = action.slots.active

        layout.prop(slot, "name_display", text="Name")

        # Draw the ID type of the slot.
        try:
            enum_items = slot.bl_rna.properties['target_id_type'].enum_items
            idtype_label = enum_items[slot.target_id_type].name
        except (KeyError, IndexError, AttributeError) as ex:
            idtype_label = str(ex)

        split = layout.split(factor=0.4)
        split.alignment = 'RIGHT'
        split.label(text="Type")
        split.alignment = 'LEFT'

        split.label(text=idtype_label, icon_value=slot.target_id_type_icon)


#######################################
# Grease Pencil Editing

class DOPESHEET_MT_gpencil_channel(Menu):
    bl_label = "Channel"

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = 'INVOKE_REGION_CHANNELS'

        layout.operator("anim.channels_delete")

        layout.separator()
        layout.operator("anim.channels_setting_toggle")
        layout.operator("anim.channels_setting_enable")
        layout.operator("anim.channels_setting_disable")

        layout.separator()
        layout.operator("anim.channels_editable_toggle")

        # XXX: to be enabled when these are ready for use!
        # layout.separator()
        # layout.operator("anim.channels_expand")
        # layout.operator("anim.channels_collapse")

        layout.separator()
        layout.operator_menu_enum("anim.channels_move", "direction", text="Move...")

        layout.separator()
        layout.operator("anim.channels_view_selected")


class DOPESHEET_MT_delete(Menu):
    bl_label = "Delete"

    def draw(self, _context):
        layout = self.layout

        layout.operator("action.delete")

        layout.separator()

        layout.operator("action.clean").channels = False
        layout.operator("action.clean", text="Clean Channels").channels = True


class DOPESHEET_MT_context_menu(Menu):
    bl_label = "Dope Sheet"

    def draw(self, context):
        layout = self.layout
        st = context.space_data

        layout.operator_context = 'INVOKE_DEFAULT'

        layout.operator("action.copy", text="Copy", icon='COPYDOWN')
        layout.operator("action.paste", text="Paste", icon='PASTEDOWN')
        layout.operator("action.paste", text="Paste Flipped", icon='PASTEFLIPDOWN').flipped = True

        layout.separator()

        layout.operator_menu_enum("action.keyframe_type", "type", text="Keyframe Type")

        if st.mode != 'GPENCIL':
            layout.operator_menu_enum("action.handle_type", "type", text="Handle Type")
            layout.operator_menu_enum("action.interpolation_type", "type", text="Interpolation Mode")
            layout.operator_menu_enum("action.easing_type", "type", text="Easing Mode")

        layout.separator()

        layout.operator("action.keyframe_insert").type = 'SEL'
        layout.operator("action.duplicate_move")

        if st.mode == 'GPENCIL':
            layout.separator()
            layout.operator("grease_pencil.delete_breakdown")

        layout.operator_context = 'EXEC_REGION_WIN'
        layout.operator("action.delete")

        layout.separator()

        layout.operator_menu_enum("action.mirror", "type", text="Mirror")
        layout.operator_menu_enum("action.snap", "type", text="Snap")


class DOPESHEET_MT_channel_context_menu(Menu):
    bl_label = "Channel"

    def draw(self, context):
        layout = self.layout

        # This menu is used from the graph editor too.
        is_graph_editor = context.area.type == 'GRAPH_EDITOR'

        layout.operator_context = 'INVOKE_REGION_CHANNELS'

        layout.separator()
        layout.operator("anim.channels_view_selected")

        layout.operator("anim.channels_setting_enable", text="Mute Channels").type = 'MUTE'
        layout.operator("anim.channels_setting_disable", text="Unmute Channels").type = 'MUTE'
        layout.separator()
        layout.operator("anim.channels_setting_enable", text="Protect Channels").type = 'PROTECT'
        layout.operator("anim.channels_setting_disable", text="Unprotect Channels").type = 'PROTECT'

        layout.separator()
        layout.operator("anim.channels_group")
        layout.operator("anim.channels_ungroup")

        layout.separator()
        layout.operator("anim.channels_editable_toggle")

        if is_graph_editor:
            operator = "graph.extrapolation_type"
        else:
            operator = "action.extrapolation_type"
        layout.operator_menu_enum(operator, "type", text="Extrapolation Mode")

        if is_graph_editor:
            layout.operator_menu_enum("graph.fmodifier_add", "type", text="Add F-Curve Modifier").only_active = False
            layout.separator()
            layout.operator("graph.hide", text="Hide Selected Curves").unselected = False
            layout.operator("graph.hide", text="Hide Unselected Curves").unselected = True
            layout.operator("graph.reveal")

        layout.separator()
        layout.operator("anim.channels_expand")
        layout.operator("anim.channels_collapse")

        layout.separator()
        layout.operator_menu_enum("anim.channels_move", "direction", text="Move...")

        layout.separator()

        layout.operator("anim.channels_delete")

        if is_graph_editor and context.space_data.mode == 'DRIVERS':
            layout.operator("graph.driver_delete_invalid")


class DOPESHEET_MT_snap_pie(Menu):
    bl_label = "Snap"

    def draw(self, _context):
        layout = self.layout
        pie = layout.menu_pie()

        pie.operator("action.snap", text="Selection to Current Frame").type = 'CFRA'
        pie.operator("action.snap", text="Selection to Nearest Frame").type = 'NEAREST_FRAME'
        pie.operator("action.snap", text="Selection to Nearest Second").type = 'NEAREST_SECOND'
        pie.operator("action.snap", text="Selection to Nearest Marker").type = 'NEAREST_MARKER'


class GreasePencilLayersDopeSheetPanel:
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'UI'
    bl_category = "View"

    @classmethod
    def poll(cls, context):
        st = context.space_data
        ob = context.object
        if st.mode != 'GPENCIL' or ob is None or ob.type != 'GREASEPENCIL':
            return False

        grease_pencil = ob.data
        active_layer = grease_pencil.layers.active
        if active_layer:
            return True

        return False


class DOPESHEET_PT_grease_pencil_mode(GreasePencilLayersDopeSheetPanel, Panel):
    bl_label = "Layer"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        ob = context.object
        grease_pencil = ob.data
        active_layer = grease_pencil.layers.active

        if active_layer:
            row = layout.row(align=True)
            row.prop(active_layer, "blend_mode", text="Blend")

            row = layout.row(align=True)
            row.prop(active_layer, "opacity", text="Opacity", slider=True)

            row = layout.row(align=True)
            row.prop(active_layer, "use_lights", text="Lights")


class DOPESHEET_PT_grease_pencil_layer_masks(GreasePencilLayersDopeSheetPanel, GreasePencil_LayerMaskPanel, Panel):
    bl_label = "Masks"
    bl_parent_id = "DOPESHEET_PT_grease_pencil_mode"
    bl_options = {'DEFAULT_CLOSED'}


class DOPESHEET_PT_grease_pencil_layer_transform(
        GreasePencilLayersDopeSheetPanel,
        GreasePencil_LayerTransformPanel,
        Panel,
):
    bl_label = "Transform"
    bl_parent_id = "DOPESHEET_PT_grease_pencil_mode"
    bl_options = {'DEFAULT_CLOSED'}


class DOPESHEET_PT_grease_pencil_layer_relations(
        GreasePencilLayersDopeSheetPanel,
        GreasePencil_LayerRelationsPanel,
        Panel,
):
    bl_label = "Relations"
    bl_parent_id = "DOPESHEET_PT_grease_pencil_mode"
    bl_options = {'DEFAULT_CLOSED'}


class DOPESHEET_PT_grease_pencil_layer_adjustments(
        GreasePencilLayersDopeSheetPanel,
        GreasePencil_LayerAdjustmentsPanel,
        Panel,
):
    bl_label = "Adjustments"
    bl_parent_id = "DOPESHEET_PT_grease_pencil_mode"
    bl_options = {'DEFAULT_CLOSED'}


class DOPESHEET_PT_grease_pencil_layer_display(
        GreasePencilLayersDopeSheetPanel,
        GreasePencil_LayerDisplayPanel,
        Panel,
):
    bl_label = "Display"
    bl_parent_id = "DOPESHEET_PT_grease_pencil_mode"
    bl_options = {'DEFAULT_CLOSED'}


class DOPESHEET_PT_ShapeKey(Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'UI'
    bl_category = "Shape Key"
    bl_label = "Shape Key"

    @classmethod
    def poll(cls, context):
        st = context.space_data
        if st.mode != 'SHAPEKEY':
            return False

        ob = context.object
        if ob is None or ob.active_shape_key is None:
            return False

        if not ob.data.shape_keys.use_relative:
            return False

        return ob.active_shape_key_index > 0

    def draw(self, context):
        draw_shape_key_properties(context, self.layout)


class DOPESHEET_PT_overlay(Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'HEADER'
    bl_label = "Overlays"
    bl_ui_units_x = 13

    def draw(self, _context):
        pass


class DOPESHEET_PT_dopesheet_overlay(Panel):
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'HEADER'
    bl_parent_id = "DOPESHEET_PT_overlay"
    bl_label = "Dope Sheet Overlays"

    def draw(self, context):
        st = context.space_data
        overlay_settings = st.overlays
        layout = self.layout

        layout.active = overlay_settings.show_overlays
        row = layout.row()
        row.active = context.workspace.use_scene_time_sync
        row.prop(overlay_settings, "show_scene_strip_range")


classes = (
    DOPESHEET_HT_header,
    DOPESHEET_HT_playback_controls,
    DOPESHEET_PT_proportional_edit,
    DOPESHEET_MT_editor_menus,
    DOPESHEET_MT_view,
    DOPESHEET_MT_cache,
    DOPESHEET_MT_select,
    DOPESHEET_MT_marker,
    DOPESHEET_MT_channel,
    DOPESHEET_MT_action,
    DOPESHEET_MT_key,
    DOPESHEET_MT_key_transform,
    DOPESHEET_MT_gpencil_channel,
    DOPESHEET_MT_delete,
    DOPESHEET_MT_context_menu,
    DOPESHEET_MT_channel_context_menu,
    DOPESHEET_MT_snap_pie,
    DOPESHEET_MT_view_pie,
    DOPESHEET_PT_filters,
    DOPESHEET_PT_action,
    DOPESHEET_PT_action_slot,
    DOPESHEET_PT_custom_props_action,
    DOPESHEET_PT_snapping,
    DOPESHEET_PT_grease_pencil_mode,
    DOPESHEET_PT_grease_pencil_layer_masks,
    DOPESHEET_PT_grease_pencil_layer_transform,
    DOPESHEET_PT_grease_pencil_layer_adjustments,
    DOPESHEET_PT_grease_pencil_layer_relations,
    DOPESHEET_PT_grease_pencil_layer_display,
    DOPESHEET_PT_ShapeKey,

    DOPESHEET_PT_overlay,
    DOPESHEET_PT_dopesheet_overlay,
)

if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
